@@ -50,7 +50,7 @@ impl Container {
             .slices()
             .first()
             .map(|s| s.header().record_counter())
-            .expect("no slices is in bulder");
+            .expect("no slices in builder");
 
         for slice in data_container.slices() {
             let slice_header = slice.header();
@@ -65,14 +65,17 @@ impl Container {
                 container_reference_sequence_id = Some(slice.header().reference_sequence_id());
             }
 
-            container_alignment_start =
-                cmp::min(container_alignment_start, slice_header.alignment_start());
+            let slice_alignment_start = slice_header
+                .alignment_start()
+                .map(i32::from)
+                .unwrap_or_default();
 
-            let slice_alignment_end =
-                slice_header.alignment_start() + slice_header.alignment_span() - 1;
+            container_alignment_start = cmp::min(container_alignment_start, slice_alignment_start);
+
+            let slice_alignment_end = slice_alignment_start + slice_header.alignment_span() - 1;
             container_alignment_end = cmp::max(container_alignment_end, slice_alignment_end);
 
-            container_record_count += slice_header.record_count();
+            container_record_count += slice_header.record_count() as Itf8;
 
             let mut slice_len = 0;
 
@@ -104,6 +107,8 @@ impl Container {
         let len = blocks.iter().map(|b| b.len() as i32).sum();
 
         let container_alignment_span = container_alignment_end - container_alignment_start + 1;
+        let container_alignment_start = sam::record::Position::try_from(container_alignment_start)
+            .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))?;
 
         let header = Header::builder()
             .set_length(len)
@@ -231,7 +236,6 @@ mod tests {
         let expected_header = Header::builder()
             .set_length(65)
             .set_reference_sequence_id(ReferenceSequenceId::None)
-            .set_start_position(0)
             .set_alignment_span(0)
             .set_record_count(0)
             .set_record_counter(0)
