@@ -18,7 +18,7 @@ use crate::{
     },
     huffman::CanonicalHuffmanDecoder,
     num::Itf8,
-    record::{self, feature, tag, Feature, ReadGroupId, Tag},
+    record::{feature, tag, Feature, Flags, NextMateFlags, ReadGroupId, Tag},
     BitReader, Record,
 };
 
@@ -46,29 +46,29 @@ impl fmt::Display for ReadRecordError {
     }
 }
 
-pub struct Reader<'a, R, S>
+pub struct Reader<'a, CDR, EDR>
 where
-    R: Read,
-    S: BufRead,
+    CDR: Read,
+    EDR: BufRead,
 {
     compression_header: &'a CompressionHeader,
-    core_data_reader: BitReader<R>,
-    external_data_readers: HashMap<Itf8, S>,
+    core_data_reader: BitReader<CDR>,
+    external_data_readers: HashMap<Itf8, EDR>,
     reference_sequence_id: ReferenceSequenceId,
-    prev_alignment_start: Itf8,
+    prev_alignment_start: Option<sam::record::Position>,
 }
 
-impl<'a, R, S> Reader<'a, R, S>
+impl<'a, CDR, EDR> Reader<'a, CDR, EDR>
 where
-    R: Read,
-    S: BufRead,
+    CDR: Read,
+    EDR: BufRead,
 {
     pub fn new(
         compression_header: &'a CompressionHeader,
-        core_data_reader: BitReader<R>,
-        external_data_readers: HashMap<Itf8, S>,
+        core_data_reader: BitReader<CDR>,
+        external_data_readers: HashMap<Itf8, EDR>,
         reference_sequence_id: ReferenceSequenceId,
-        initial_alignment_start: Itf8,
+        initial_alignment_start: Option<sam::record::Position>,
     ) -> Self {
         Self {
             compression_header,
@@ -79,45 +79,34 @@ where
         }
     }
 
-    pub fn read_record(&mut self, record: &mut Record) -> io::Result<()> {
-        record.bam_bit_flags = self
-            .read_bam_bit_flags()
-            .and_then(|n| {
-                u16::try_from(n).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
-            })
-            .map(sam::record::Flags::from)?;
+    pub fn read_record(&mut self) -> io::Result<Record> {
+        let bam_bit_flags = self.read_bam_bit_flags()?;
+        let cram_bit_flags = self.read_cram_bit_flags()?;
 
-        record.cram_bit_flags = self
-            .read_cram_bit_flags()
-            .and_then(|n| {
-                u8::try_from(n).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
-            })
-            .map(record::Flags::from)?;
+        let mut record = Record {
+            bam_bit_flags,
+            cram_bit_flags,
+            ..Default::default()
+        };
 
-        self.read_positional_data(record)?;
+        self.read_positional_data(&mut record)?;
+        self.read_read_names(&mut record)?;
+        self.read_mate_data(&mut record)?;
 
-        let preservation_map = self.compression_header.preservation_map();
-
-        // Missing read names are generated when resolving mates.
-        if preservation_map.read_names_included() {
-            record.read_name = self.read_read_name()?;
-        }
-
-        self.read_mate_data(record)?;
-        self.read_tag_data(record)?;
+        record.tags = self.read_tag_data()?;
 
         if record.bam_flags().is_unmapped() {
-            self.read_unmapped_read(record)?;
+            self.read_unmapped_read(&mut record)?;
         } else {
-            self.read_mapped_read(record)?;
+            self.read_mapped_read(&mut record)?;
         }
 
         self.prev_alignment_start = record.alignment_start();
 
-        Ok(())
+        Ok(record)
     }
 
-    fn read_bam_bit_flags(&mut self) -> io::Result<Itf8> {
+    fn read_bam_bit_flags(&mut self) -> io::Result<sam::record::Flags> {
         let encoding = self
             .compression_header
             .data_series_encoding_map()
@@ -128,9 +117,11 @@ where
             &mut self.core_data_reader,
             &mut self.external_data_readers,
         )
+        .and_then(|n| u16::try_from(n).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e)))
+        .map(sam::record::Flags::from)
     }
 
-    fn read_cram_bit_flags(&mut self) -> io::Result<Itf8> {
+    fn read_cram_bit_flags(&mut self) -> io::Result<Flags> {
         let encoding = self
             .compression_header
             .data_series_encoding_map()
@@ -141,6 +132,8 @@ where
             &mut self.core_data_reader,
             &mut self.external_data_readers,
         )
+        .and_then(|n| u8::try_from(n).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e)))
+        .map(Flags::from)
     }
 
     fn read_positional_data(&mut self, record: &mut Record) -> io::Result<()> {
@@ -159,24 +152,9 @@ where
                     .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))?
             };
 
-        record.read_length = self.read_read_length().and_then(|n| {
-            usize::try_from(n).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
-        })?;
-
-        let ap_data_series_delta = self
-            .compression_header
-            .preservation_map()
-            .ap_data_series_delta();
-
-        let alignment_start = self.read_alignment_start()?;
-
-        record.alignment_start = if ap_data_series_delta {
-            self.prev_alignment_start + alignment_start
-        } else {
-            alignment_start
-        };
-
-        record.read_group = self.read_read_group().map(ReadGroupId::from)?;
+        record.read_length = self.read_read_length()?;
+        record.alignment_start = self.read_alignment_start()?;
+        record.read_group = self.read_read_group()?;
 
         Ok(())
     }
@@ -200,7 +178,7 @@ where
             })
     }
 
-    fn read_read_length(&mut self) -> io::Result<Itf8> {
+    fn read_read_length(&mut self) -> io::Result<usize> {
         let encoding = self
             .compression_header
             .data_series_encoding_map()
@@ -211,22 +189,42 @@ where
             &mut self.core_data_reader,
             &mut self.external_data_readers,
         )
+        .and_then(|n| usize::try_from(n).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e)))
     }
 
-    fn read_alignment_start(&mut self) -> io::Result<Itf8> {
+    fn read_alignment_start(&mut self) -> io::Result<Option<sam::record::Position>> {
+        let ap_data_series_delta = self
+            .compression_header
+            .preservation_map()
+            .ap_data_series_delta();
+
         let encoding = self
             .compression_header
             .data_series_encoding_map()
             .in_seq_positions_encoding();
 
-        decode_itf8(
+        let mut alignment_start = decode_itf8(
             encoding,
             &mut self.core_data_reader,
             &mut self.external_data_readers,
-        )
+        )?;
+
+        if ap_data_series_delta {
+            let prev_alignment_start = self.prev_alignment_start.map(i32::from).unwrap_or_default();
+            let delta = alignment_start;
+            alignment_start = prev_alignment_start + delta;
+        }
+
+        if alignment_start == 0 {
+            Ok(None)
+        } else {
+            sam::record::Position::try_from(alignment_start)
+                .map(Some)
+                .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
+        }
     }
 
-    fn read_read_group(&mut self) -> io::Result<Itf8> {
+    fn read_read_group(&mut self) -> io::Result<ReadGroupId> {
         let encoding = self
             .compression_header
             .data_series_encoding_map()
@@ -237,6 +235,18 @@ where
             &mut self.core_data_reader,
             &mut self.external_data_readers,
         )
+        .map(ReadGroupId::from)
+    }
+
+    fn read_read_names(&mut self, record: &mut Record) -> io::Result<()> {
+        let preservation_map = self.compression_header.preservation_map();
+
+        // Missing read names are generated when resolving mates.
+        if preservation_map.read_names_included() {
+            record.read_name = self.read_read_name()?;
+        }
+
+        Ok(())
     }
 
     fn read_read_name(&mut self) -> io::Result<Vec<u8>> {
@@ -263,12 +273,8 @@ where
         let flags = record.flags();
 
         if flags.is_detached() {
-            record.next_mate_bit_flags = self
-                .read_next_mate_bit_flags()
-                .and_then(|n| {
-                    u8::try_from(n).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
-                })
-                .map(record::NextMateFlags::from)?;
+            record.next_mate_bit_flags = self.read_next_mate_bit_flags()?;
+
             let next_mate_flags = record.next_mate_flags();
 
             if next_mate_flags.is_on_negative_strand() {
@@ -285,17 +291,8 @@ where
                 record.read_name = self.read_read_name()?;
             }
 
-            record.next_fragment_reference_sequence_id = self
-                .read_next_fragment_reference_sequence_id()
-                .and_then(|id| {
-                    if id == bam::record::reference_sequence_id::UNMAPPED {
-                        Ok(None)
-                    } else {
-                        bam::record::ReferenceSequenceId::try_from(id)
-                            .map(Some)
-                            .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
-                    }
-                })?;
+            record.next_fragment_reference_sequence_id =
+                self.read_next_fragment_reference_sequence_id()?;
 
             record.next_mate_alignment_start = self.read_next_mate_alignment_start()?;
             record.template_size = self.read_template_size()?;
@@ -306,8 +303,9 @@ where
         Ok(())
     }
 
-    fn read_next_mate_bit_flags(&mut self) -> io::Result<Itf8> {
-        self.compression_header
+    fn read_next_mate_bit_flags(&mut self) -> io::Result<NextMateFlags> {
+        let encoding = self
+            .compression_header
             .data_series_encoding_map()
             .next_mate_bit_flags_encoding()
             .ok_or_else(|| {
@@ -315,18 +313,22 @@ where
                     io::ErrorKind::InvalidData,
                     ReadRecordError::MissingDataSeriesEncoding(DataSeries::NextMateBitFlags),
                 )
-            })
-            .and_then(|encoding| {
-                decode_itf8(
-                    encoding,
-                    &mut self.core_data_reader,
-                    &mut self.external_data_readers,
-                )
-            })
+            })?;
+
+        decode_itf8(
+            encoding,
+            &mut self.core_data_reader,
+            &mut self.external_data_readers,
+        )
+        .and_then(|n| u8::try_from(n).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e)))
+        .map(NextMateFlags::from)
     }
 
-    fn read_next_fragment_reference_sequence_id(&mut self) -> io::Result<Itf8> {
-        self.compression_header
+    fn read_next_fragment_reference_sequence_id(
+        &mut self,
+    ) -> io::Result<Option<bam::record::ReferenceSequenceId>> {
+        let encoding = self
+            .compression_header
             .data_series_encoding_map()
             .next_fragment_reference_sequence_id_encoding()
             .ok_or_else(|| {
@@ -336,18 +338,27 @@ where
                         DataSeries::NextFragmentReferenceSequenceId,
                     ),
                 )
-            })
-            .and_then(|encoding| {
-                decode_itf8(
-                    encoding,
-                    &mut self.core_data_reader,
-                    &mut self.external_data_readers,
-                )
-            })
+            })?;
+
+        decode_itf8(
+            encoding,
+            &mut self.core_data_reader,
+            &mut self.external_data_readers,
+        )
+        .and_then(|id| {
+            if id == bam::record::reference_sequence_id::UNMAPPED {
+                Ok(None)
+            } else {
+                bam::record::ReferenceSequenceId::try_from(id)
+                    .map(Some)
+                    .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
+            }
+        })
     }
 
-    fn read_next_mate_alignment_start(&mut self) -> io::Result<Itf8> {
-        self.compression_header
+    fn read_next_mate_alignment_start(&mut self) -> io::Result<Option<sam::record::Position>> {
+        let encoding = self
+            .compression_header
             .data_series_encoding_map()
             .next_mate_alignment_start_encoding()
             .ok_or_else(|| {
@@ -355,14 +366,22 @@ where
                     io::ErrorKind::InvalidData,
                     ReadRecordError::MissingDataSeriesEncoding(DataSeries::NextMateAlignmentStart),
                 )
-            })
-            .and_then(|encoding| {
-                decode_itf8(
-                    encoding,
-                    &mut self.core_data_reader,
-                    &mut self.external_data_readers,
-                )
-            })
+            })?;
+
+        decode_itf8(
+            encoding,
+            &mut self.core_data_reader,
+            &mut self.external_data_readers,
+        )
+        .and_then(|n| {
+            if n == 0 {
+                Ok(None)
+            } else {
+                sam::record::Position::try_from(n)
+                    .map(Some)
+                    .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
+            }
+        })
     }
 
     fn read_template_size(&mut self) -> io::Result<Itf8> {
@@ -403,16 +422,23 @@ where
             })
     }
 
-    fn read_tag_data(&mut self, record: &mut Record) -> io::Result<()> {
-        let tag_line = self.read_tag_line()?;
+    fn read_tag_data(&mut self) -> io::Result<Vec<Tag>> {
+        use bam::reader::record::data::field::read_value;
 
-        let preservation_map = self.compression_header.preservation_map();
-        let tag_ids_dictionary = preservation_map.tag_ids_dictionary();
-        let tag_keys = &tag_ids_dictionary[tag_line as usize];
+        let tag_line = self.read_tag_line().and_then(|i| {
+            usize::try_from(i).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
+        })?;
+
+        let tag_keys = self
+            .compression_header
+            .preservation_map()
+            .tag_ids_dictionary()
+            .get(tag_line)
+            .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "invalid tag line"))?;
 
         let tag_encoding_map = self.compression_header.tag_encoding_map();
 
-        record.tags.clear();
+        let mut tags = Vec::with_capacity(tag_keys.len());
 
         for key in tag_keys {
             let id = key.id();
@@ -430,14 +456,14 @@ where
                 None,
             )?;
 
-            let mut data_reader = bam::record::data::Reader::new(&data[..]);
-            let value = data_reader.read_value_type(key.ty())?;
+            let mut data_reader = &data[..];
+            let value = read_value(&mut data_reader, key.ty())?;
 
             let tag = Tag::new(*key, value);
-            record.add_tag(tag);
+            tags.push(tag);
         }
 
-        Ok(())
+        Ok(tags)
     }
 
     fn read_tag_line(&mut self) -> io::Result<Itf8> {
@@ -466,9 +492,7 @@ where
             record.add_feature(feature);
         }
 
-        record.mapping_quality = self
-            .read_mapping_quality()
-            .map(|n| sam::record::MappingQuality::from(n as u8))?;
+        record.mapping_quality = self.read_mapping_quality()?;
 
         let flags = record.flags();
 
@@ -507,71 +531,69 @@ where
     }
 
     fn read_feature(&mut self, prev_position: i32) -> io::Result<Feature> {
-        let code = self
-            .read_feature_code()
-            .map(|id| id as u8 as char)
-            .and_then(|id| {
-                feature::Code::try_from(id)
-                    .map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
-            })?;
+        use feature::Code;
 
-        let position = prev_position + self.read_feature_position()?;
+        let code = self.read_feature_code()?;
+
+        let delta = self.read_feature_position()?;
+        let position = prev_position + delta;
 
         match code {
-            feature::Code::Bases => {
+            Code::Bases => {
                 let bases = self.read_stretches_of_bases()?;
                 Ok(Feature::Bases(position, bases))
             }
-            feature::Code::Scores => {
+            Code::Scores => {
                 let quality_scores = self.read_stretches_of_quality_scores()?;
                 Ok(Feature::Scores(position, quality_scores))
             }
-            feature::Code::ReadBase => {
+            Code::ReadBase => {
                 let base = self.read_base()?;
                 let quality_score = self.read_quality_score()?;
                 Ok(Feature::ReadBase(position, base, quality_score))
             }
-            feature::Code::Substitution => {
+            Code::Substitution => {
                 let code = self.read_base_substitution_code()?;
                 Ok(Feature::Substitution(position, code))
             }
-            feature::Code::Insertion => {
+            Code::Insertion => {
                 let bases = self.read_insertion()?;
                 Ok(Feature::Insertion(position, bases))
             }
-            feature::Code::Deletion => {
+            Code::Deletion => {
                 let len = self.read_deletion_length()?;
                 Ok(Feature::Deletion(position, len))
             }
-            feature::Code::InsertBase => {
+            Code::InsertBase => {
                 let base = self.read_base()?;
                 Ok(Feature::InsertBase(position, base))
             }
-            feature::Code::QualityScore => {
+            Code::QualityScore => {
                 let score = self.read_quality_score()?;
                 Ok(Feature::QualityScore(position, score))
             }
-            feature::Code::ReferenceSkip => {
+            Code::ReferenceSkip => {
                 let len = self.read_reference_skip_length()?;
                 Ok(Feature::ReferenceSkip(position, len))
             }
-            feature::Code::SoftClip => {
+            Code::SoftClip => {
                 let bases = self.read_soft_clip()?;
                 Ok(Feature::SoftClip(position, bases))
             }
-            feature::Code::Padding => {
+            Code::Padding => {
                 let len = self.read_padding()?;
                 Ok(Feature::Padding(position, len))
             }
-            feature::Code::HardClip => {
+            Code::HardClip => {
                 let len = self.read_hard_clip()?;
                 Ok(Feature::HardClip(position, len))
             }
         }
     }
 
-    fn read_feature_code(&mut self) -> io::Result<Itf8> {
-        self.compression_header
+    fn read_feature_code(&mut self) -> io::Result<feature::Code> {
+        let encoding = self
+            .compression_header
             .data_series_encoding_map()
             .read_features_codes_encoding()
             .ok_or_else(|| {
@@ -579,14 +601,17 @@ where
                     io::ErrorKind::InvalidData,
                     ReadRecordError::MissingDataSeriesEncoding(DataSeries::ReadFeaturesCodes),
                 )
-            })
-            .and_then(|encoding| {
-                decode_itf8(
-                    encoding,
-                    &mut self.core_data_reader,
-                    &mut self.external_data_readers,
-                )
-            })
+            })?;
+
+        decode_itf8(
+            encoding,
+            &mut self.core_data_reader,
+            &mut self.external_data_readers,
+        )
+        .map(|id| id as u8 as char)
+        .and_then(|id| {
+            feature::Code::try_from(id).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e))
+        })
     }
 
     fn read_feature_position(&mut self) -> io::Result<Itf8> {
@@ -823,8 +848,9 @@ where
             })
     }
 
-    fn read_mapping_quality(&mut self) -> io::Result<Itf8> {
-        self.compression_header
+    fn read_mapping_quality(&mut self) -> io::Result<sam::record::MappingQuality> {
+        let encoding = self
+            .compression_header
             .data_series_encoding_map()
             .mapping_qualities_encoding()
             .ok_or_else(|| {
@@ -832,14 +858,15 @@ where
                     io::ErrorKind::InvalidData,
                     ReadRecordError::MissingDataSeriesEncoding(DataSeries::MappingQualities),
                 )
-            })
-            .and_then(|encoding| {
-                decode_itf8(
-                    encoding,
-                    &mut self.core_data_reader,
-                    &mut self.external_data_readers,
-                )
-            })
+            })?;
+
+        decode_itf8(
+            encoding,
+            &mut self.core_data_reader,
+            &mut self.external_data_readers,
+        )
+        .and_then(|n| u8::try_from(n).map_err(|e| io::Error::new(io::ErrorKind::InvalidData, e)))
+        .map(sam::record::MappingQuality::from)
     }
 
     fn read_unmapped_read(&mut self, record: &mut Record) -> io::Result<()> {
@@ -869,14 +896,14 @@ where
     }
 }
 
-fn decode_byte<R, S>(
+fn decode_byte<CDR, EDR>(
     encoding: &Encoding,
-    core_data_reader: &mut BitReader<R>,
-    external_data_readers: &mut HashMap<Itf8, S>,
+    core_data_reader: &mut BitReader<CDR>,
+    external_data_readers: &mut HashMap<Itf8, EDR>,
 ) -> io::Result<u8>
 where
-    R: Read,
-    S: Read,
+    CDR: Read,
+    EDR: Read,
 {
     match encoding {
         Encoding::External(block_content_id) => {
@@ -893,23 +920,23 @@ where
         }
         Encoding::Huffman(alphabet, bit_lens) => {
             let decoder = CanonicalHuffmanDecoder::new(alphabet, bit_lens);
-            decoder.read(core_data_reader).map(|i| i as u8)
+            decoder.decode(core_data_reader).map(|i| i as u8)
         }
         Encoding::Beta(offset, len) => core_data_reader
-            .read_u32(*len as usize)
+            .read_u32(*len)
             .map(|i| (i as i32 - offset) as u8),
         _ => todo!("decode_byte: {:?}", encoding),
     }
 }
 
-fn decode_itf8<R, S>(
+fn decode_itf8<CDR, EDR>(
     encoding: &Encoding,
-    core_data_reader: &mut BitReader<R>,
-    external_data_readers: &mut HashMap<Itf8, S>,
+    core_data_reader: &mut BitReader<CDR>,
+    external_data_readers: &mut HashMap<Itf8, EDR>,
 ) -> io::Result<Itf8>
 where
-    R: Read,
-    S: Read,
+    CDR: Read,
+    EDR: Read,
 {
     match encoding {
         Encoding::External(block_content_id) => {
@@ -926,24 +953,22 @@ where
         }
         Encoding::Huffman(alphabet, bit_lens) => {
             let decoder = CanonicalHuffmanDecoder::new(alphabet, bit_lens);
-            decoder.read(core_data_reader)
+            decoder.decode(core_data_reader)
         }
-        Encoding::Beta(offset, len) => core_data_reader
-            .read_u32(*len as usize)
-            .map(|i| (i as i32 - offset)),
+        Encoding::Beta(offset, len) => core_data_reader.read_u32(*len).map(|i| (i as i32 - offset)),
         _ => todo!("decode_itf8: {:?}", encoding),
     }
 }
 
-fn decode_byte_array<R, S>(
+fn decode_byte_array<CDR, EDR>(
     encoding: &Encoding,
-    core_data_reader: &mut BitReader<R>,
-    external_data_readers: &mut HashMap<Itf8, S>,
+    core_data_reader: &mut BitReader<CDR>,
+    external_data_readers: &mut HashMap<Itf8, EDR>,
     buf: Option<Vec<u8>>,
 ) -> io::Result<Vec<u8>>
 where
-    R: Read,
-    S: BufRead,
+    CDR: Read,
+    EDR: BufRead,
 {
     match encoding {
         Encoding::External(block_content_id) => {
